import asyncio
import gzip
import pathlib
import urllib

import aiohttp
from tqdm.asyncio import tqdm_asyncio


BASE_ALPH = tuple("123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")
BASE_DICT = dict((c, v) for v, c in enumerate(BASE_ALPH))
BASE_LEN = len(BASE_ALPH)


def urljoin(*args: str) -> str:
    """Safely join URLs.

    Returns:
        str: The joined URL.
    """
    return "/".join(map(lambda x: str(x).rstrip("/"), args))


def base61_decode(string: str) -> int:
    """ Decode a base61 string.

    Args:
        string (str): The string to decode.

    Returns:
        str: The decoded string.
    """
    num = 0
    for char in string:
        num = num * BASE_LEN + BASE_DICT[char]
    return num


def base61_encode(num: int) -> str:
    """ Encode a number in base61.

    Args:
        num (int): The number to encode.

    Returns:
        str: The encoded string.
    """

    if not num:
        return BASE_ALPH[0]

    encoding = ""
    while num:
        num, rem = divmod(num, BASE_LEN)
        encoding = BASE_ALPH[rem] + encoding
    return encoding


def get_file_size(filename: str) -> float:
    """ This function returns the size of a file in MB
    
    Args:
        filename (str): The path to the file

    Returns:
        float: The size of the file in MB
    """
    mbhr = pathlib.Path(filename).stat().st_size / (1024 ** 2)
    return round(mbhr, 4)


async def download_file(url: str) -> bytes:
    """ Download a file from a URL.

    Args:
        url (str): The URL to download from.
    
    Returns:
        bytes: The content of the file.
    """    
    # Set semaphores
    limit = asyncio.Semaphore(5)
    session_timeout = aiohttp.ClientTimeout(total=None, sock_connect=1500, sock_read=1500)
    async with limit:        
        async with aiohttp.ClientSession(timeout=session_timeout) as session:
            async with session.get(url) as resp:
                content = await resp.read()
                return content


async def write_to_file(content: bytes, filename: pathlib.Path) -> None:
    """ Download a file from a URL.

    Args:
        content (bytes): The content to write.
        filename (pathlib.Path): The filename to save to.
    """
    with open(filename, "wb") as outfile:
        outfile.write(content)


async def extract_file(in_filename: pathlib.Path, out_filename: pathlib.Path) -> int:
    """ Extract a file from a gzip archive.

    Args:
        in_filename (pathlib.Path): The zipped filename.
        out_filename (pathlib.Path): The unzipped filename.
    
    Returns:
        int: The number of bytes extracted.
    """
    sum_bytes = 0
    with gzip.open(in_filename, "rb") as infile:
        with open(out_filename, "wb") as outfile:
            for line in infile:
                outfile.write(line)
                sum_bytes += len(line)

    # Remove the .zip file
    in_filename.unlink()

    return sum_bytes


async def download_and_extract_task(url: str, path: pathlib.Path) -> None:
    """ Create a task to download and extract a file.

    Args:
        url (str): The URL to download from.
        out_filename (pathlib.Path): The filename to save to.
    """

    content = await download_file(url)

    # get the name
    gzip_name = path / pathlib.Path(url).name
    out_filename = path / gzip_name.parent / gzip_name.stem    
    await write_to_file(content, gzip_name)
    await extract_file(gzip_name, out_filename)


async def _download_and_extract(
    uri_files: str, path: int, bar_msg: str, timebar: int
) -> None:
    """ Create a list of tasks to download and extract a file.

    Args:
        url (str): The URL to download from.
        nfiles (int): The number of files to download.
        out_dir (pathlib.Path): The directory to save the files to.
    """

    # Get the file names
    tasks = []
    for index in range(timebar):
        if index <= (timebar - len(uri_files)):
            tasks.append(asyncio.sleep(0))
        else:
            url_filename = uri_files[index + len(uri_files) - timebar]
            tasks.append(download_and_extract_task(url_filename, path))

    _ = await tqdm_asyncio.gather(*tasks, desc=bar_msg)


def download_and_extract(
    uri_files: list, path: pathlib.Path, bar_msg: str = "Downloading Files",  timebar: int = 0
) -> None:
    """ Download and extract a file."""
    asyncio.run(_download_and_extract(uri_files, path, bar_msg, timebar))


def download_split(
    ml_collection: dict,
    path: pathlib.Path,
    split: str = "train",
    force: bool = False
) -> None:
    """ Download a set of files. """
    catalog = ml_collection["ml:%s" % split].copy()

    # get all the files to download
    all_files = [base61_encode(x).zfill(7) + ".safetensors" for x in range(catalog["n_items"])]
    downloaded_files = [x.name for x in (path / split).glob("*.safetensors")]
    
    if force:
        # clear the downloaded files before downloading
        for x in (path / split).glob("*.safetensors"):
            x.unlink()        
        to_download = all_files
    else:
        if len(downloaded_files) == len(all_files):
            return None        
        # Start downloading from the last files
        start_download = len(downloaded_files) - 3
        if start_download < 0:
            start_download = 0
        to_download = all_files[start_download:]
        
    # get the urls of the files to download
    uri_files = [urljoin(catalog["url"], x + ".gz") for x in to_download]
    
    # Download the files
    params = {
        "uri_files": uri_files, "path": path / split,
        "bar_msg": "%s dataset" % split, "timebar": len(all_files)
    }
    
    try:
        download_and_extract(
            uri_files=uri_files,
            path=path / split,
            bar_msg="%s dataset" % split,
            timebar=len(all_files),
        )
    except:
        print("Restarting download ...")
        download_split(ml_collection, path, split, force=False)

    return None


def simple_download_extract(url_file: str, filename: pathlib.Path) -> None:
    """ Download a file from a URL.

    Args:
        url (str): The URL to download from.
        filename (pathlib.Path): The filename to save to.
    """
    filename2 = filename.parent / ("%s.gz" % filename.stem)

    # if the file exists remove it
    if filename.exists():
        filename.unlink()

    if filename2.exists():
        filename2.unlink()

    # batch download the file
    with urllib.request.urlopen(url_file) as r:
        for chunk in iter(lambda: r.read(4096), b""):
            with open(filename2, "ab") as f:
                f.write(chunk)

    # Extract the file
    with gzip.open(filename2, "rb") as infile:
        with open(filename, "wb") as outfile:
            for line in infile:
                outfile.write(line)

    return None



import re
from datetime import datetime
from typing import Dict, List, Literal, Optional

import pydantic

from ..catalog.datamodel import Catalog

__split_strategy__ = Literal[
    "random sampling",
    "spatial stratified random sampling",
    "spatiotemporal stratified random sampling",
    "other",
]
__cvtask__ = Literal[
    "multi-class classification",
    "binary classification",
    "object detection",
    "multi-class semantic segmentation",
    "binary semantic segmentation",
    "instance segmentation",
    "image to image",
    "multi-image to image",
    "image to multi-image",
    "multi-image to multi-image",
    "regression",
    "multi-task",
    "other",
]

__bands__ = Literal[
    "aerosol",
    "blue",
    "green1",
    "green",
    "yellow",
    "red",
    "rededge1",
    "rededge2",
    "rededge3",
    "nir",
    "nir2",
    "watervapor",
    "cirrus",
    "swir1",
    "swir2",
    "thermal1",
    "thermal2",
    "hv",
    "hh",
    "vv",
    "vh",
    "sar-derived",
    "dem",
    "dem-derived",
    "landcover",
    "landcover-derived",
    "other",
]

__sensor__ = Literal[
    "sentinel-2 msi",
    "landsat-8 oli/tirs",
    "landsat-8 oli",
    "landsat-8 tirs",
    "landsat-7 etm+",
    "landsat-7 etm+ only slc-off",
    "landsat-5 tm",
    "landsat-4 tm",
    "landsat-1 to 5 mss",
    "modis",
    "modis-derived",
    "sentinel-1 sar",
    "aerial",
    "uav",
    "aster",
    "hyperion",
    "cbers",
    "planetscope",
    "rapideye",
    "spot 1 to 5",
    "spot 6/7",
    "ikonos",
    "worldview-2",
    "worldview-3",
    "other",
]

__dtype__ = Literal[
    "bool", "uint8", "int8", "uint16", "int16", "int32", "float16", "float32", "float64"
]


class HyperLink(pydantic.BaseModel):
    link: str

    @pydantic.field_validator("link")
    @classmethod
    def check_link(cls, v):
        if not re.match(r"^https?://", v):
            raise ValueError("link must be a valid URL")
        return v


class Reviewer(pydantic.BaseModel):
    name: str
    orcid: Optional[str] = None
    score: Literal[0, 1, 2, 3, 4, 5]
    issue_link: HyperLink

    @pydantic.field_validator("orcid")
    @classmethod
    def check_orcid(cls, v):
        if v is not None:
            if not re.match(r"^https://orcid.org/", v):
                raise ValueError("orcid must be a valid URL")
        return v

    @pydantic.field_validator("score")
    @classmethod
    def check_score(cls, v):
        if v not in [0, 1, 2, 3, 4, 5]:
            msg = "%s%s%s" % (
                "score must be a integer between 0 and 5",
                " (0: very poor, 1: poor, 2: fair, 3: good)",
                " 4: very good, 5: excellent)",
            )
            raise ValueError(msg)
        return v


class Reviewers(pydantic.BaseModel):
    reviewers: Optional[List[Reviewer]] = None


class License(pydantic.BaseModel):
    name: str
    link: HyperLink


class Licenses(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid")
    licenses: List[License]
    additional_comments: Optional[str] = None

    @pydantic.field_validator("licenses")
    @classmethod
    def check_licenses(cls, v):
        if len(v) == 0:
            raise ValueError("licenses must be a list with at least one element")
        return v


class SpatialExtent(pydantic.BaseModel):
    bbox: List[float]

    @pydantic.field_validator("bbox")
    @classmethod
    def check_bbox(cls, v):
        if len(v) != 4:
            msg = "%s%s" % (
                "bbox must be a list with 4 ",
                "elements: [minx, miny, maxx, maxy]",
            )
            raise ValueError(msg)
        return v


class TemporalExtent(pydantic.BaseModel):
    interval: List[datetime]

    @pydantic.field_validator("interval")
    @classmethod
    def check_interval(cls, v):
        if len(v) != 2:
            msg = "%s%s" % (
                "interval must be a list with 2 ",
                "elements: [start_datetime, end_datetime]",
            )
            raise ValueError(msg)

        if v[0] >= v[1]:
            raise ValueError("end_datetime must be after start_datetime")

        return v


class Extent(pydantic.BaseModel):
    spatial: SpatialExtent
    temporal: TemporalExtent


class Author(pydantic.BaseModel):
    name: str
    orcid: Optional[str] = None
    organization: Optional[str] = None

    @pydantic.field_validator("orcid")
    @classmethod
    def check_orcid(cls, v):
        if v is not None:
            if not re.match(r"^https://orcid.org/", v):
                raise ValueError("orcid must be a valid URL")
        return v


class Authors(pydantic.BaseModel):
    authors: List[Author]
    additional_comments: Optional[str] = None
    curated_by: HyperLink
    reference: HyperLink


class Collection(pydantic.BaseModel):
    """ML-STAC Catalog dataclass."""

    model_config = pydantic.ConfigDict(extra="allow")

    # General information
    mlstac_version: str
    name: str
    train: Catalog
    val: Catalog
    test: Catalog

    # Additional information
    authors: Authors
    licenses: Licenses
    split_strategy: __split_strategy__
    cv_task: __cvtask__
    sensor: __sensor__
    bands: List[__bands__]
    review: Reviewers
    extent: Extent
    dtype_input: __dtype__
    dtype_target: __dtype__
    dtype_reference: __dtype__
    size: int
    keywords: List[str]
    description: str
    extra: Dict[str, str]
    issue_link: HyperLink

    @pydantic.field_validator("mlstac_version")
    @classmethod
    def check_mlstac_version(cls, v):
        regex_exp = re.compile(r"^[0-9]+.[0-9]+.[0-9]+$")
        if not regex_exp.match(v):
            msg = "%s%s" % (
                "mlstac_version must be from the ",
                "form <major>.<minor>.<patch>",
            )
            raise ValueError(msg)

    @pydantic.field_validator("name")
    @classmethod
    def check_name(cls, v):
        regex_exp = re.compile(r"^[a-zA-Z0-9]+$")
        if not regex_exp.match(v):
            raise ValueError("name must be a string with only letters and numbers")
        return v

    @pydantic.field_validator("split_strategy")
    @classmethod
    def check_split_strategy(cls, v):
        if v not in __split_strategy__.__args__:
            raise ValueError("split_strategy is not valid.")
        return v

    @pydantic.field_validator("cv_task")
    @classmethod
    def check_cv_task(cls, v):
        if v not in __cvtask__.__args__:
            raise ValueError("cv_task is not valid.")
        return v

    @pydantic.field_validator("sensor")
    @classmethod
    def check_sensor(cls, v):
        if v not in __sensor__.__args__:
            raise ValueError("sensor is not valid.")
        return v

    @pydantic.field_validator("bands")
    @classmethod
    def check_bands(cls, v):
        for band in v:
            if band not in __bands__.__args__:
                raise ValueError("bands is not valid.")
        return v

    @pydantic.field_validator("dtype_input", "dtype_target", "dtype_reference")
    @classmethod
    def check_dtype_input(cls, v):
        if v not in __dtype__.__args__:
            raise ValueError("dtype_input is not valid.")
        return v

    @pydantic.field_validator("keywords")
    @classmethod
    def check_keywords(cls, v):
        if len(v) == 0:
            raise ValueError("keywords must be a list with at least one element")
        return v

    @pydantic.field_validator("extra")
    @classmethod
    def check_extra(cls, v):
        # only integers, floats and strings bool are allowed
        to_eval = (int, float, str, bool)
        condition = all(isinstance(x, to_eval) for x in v.values())
        if not condition:
            raise ValueError("extra values must be integers, floats, strings or bools.")



from mlstac.sample.datamodel import Sample, SampleMetadata, SampleTensor
from mlstac.catalog.datamodel import Catalog
from mlstac.collection.datamodel import (
    Author,
    Authors,
    Collection,
    Extent,
    HyperLink,
    License,
    Licenses,
    Reviewer,
    Reviewers,
    SpatialExtent,
    TemporalExtent,
)
from mlstac.api.main import load, download
from mlstac.api.datasets import LocalDataset, StreamDataset

from mlstac.api.nest_asyncio import apply as nest_asyncio_apply

# Patch asyncio to make its event loop reentrant.
nest_asyncio_apply()



import pathlib

import safetensors
from mlstac.api.utils import base61_encode, simple_download_extract, urljoin


class LocalDataset:
    def __init__(
        self, path: pathlib.Path, framework: str = "torch", device: str = "cpu"
    ):
        self.path = path
        self.files = list(path.glob("*.safetensors"))
        self.framework = framework
        self.device = device

    def __getitem__(self, index):
        tensors = {}
        with safetensors.safe_open(
            filename=self.files[index], framework=self.framework, device=self.device
        ) as f:
            for k in f.keys():
                tensors[k] = f.get_tensor(k)
            return tensors

    def __len__(self):
        return len(self.files)


class StreamDataset:
    def __init__(
        self,
        url: str,
        tempfile: pathlib.Path,
        framework: str = "torch",
        device: str = "cpu",
    ):
        self.framework = framework
        self.device = device
        self.url = url
        self.tempfile = tempfile

    def __getitem__(self, index):
        file_snippet = base61_encode(index).zfill(7)

        url_file = urljoin(self.url, "%s.safetensors.gz" % file_snippet)

        simple_download_extract(url_file, self.tempfile)
        tensors = {}
        with safetensors.safe_open(
            filename=self.tempfile, framework=self.framework, device=self.device
        ) as f:
            for k in f.keys():
                tensors[k] = f.get_tensor(k)
            return tensors

    def __len__(self):
        return len(self.files)



import re
from typing import Literal

import pydantic


class Catalog(pydantic.BaseModel):
    """ML-STAC Catalog dataclass."""

    model_config = pydantic.ConfigDict(extra="forbid")

    name: Literal["train", "test", "val"]
    n_items: int

    @pydantic.field_validator("name")
    def name_must_be_valid(cls, value: str) -> str:
        """name must be only train, test, or val"""
        if not re.match(r"train|test|val", value):
            raise ValueError("name must be one of train, test, or val")
        return value

    @pydantic.field_validator("n_items")
    def n_items_must_be_valid(cls, value: int) -> int:
        """n_items must be positive"""
        # must be positive
        if value <= 0:
            raise ValueError("n_items must be positive")
        return value



import json
import pathlib
import urllib
from tempfile import NamedTemporaryFile
from typing import Union

from mlstac.api.datasets import LocalDataset, StreamDataset
from mlstac.api.utils import download_split


def load_secret(secret: Union[str, pathlib.Path, dict]) -> dict:
    """ Load a secret from a URL or a local file."""
    if isinstance(secret, str):
        if pathlib.Path(secret).exists():
            with open(secret, "r") as f:
                ml_collection = json.load(f)
        else:
            with urllib.request.urlopen(secret) as response:
                binary_data = response.read()
                ml_collection = json.loads(binary_data.decode("utf-8"))
    elif isinstance(secret, dict):
        ml_collection = secret
    else:
        raise ValueError("The secret must be a URL, a local file or a dictionary.")

    return ml_collection


def download(
    secret: Union[str, pathlib.Path, dict],
    path: Union[str, pathlib.Path],
    split: Union[str, None] = None,
    verbose: bool = True,
):
    """ Download a dataset that follows the ML-STAC specification.

    Args:
        secret (Union[str, dict]): An URI or a local path to a JSON 
        file or a dictionary that implements the ML-STAC Collection
        specification.
        path (Union[str, pathlib.Path]): The path where the dataset
        will be saved.
        split (str, optional): The split of the dataset to download.
        If None, the entire dataset will be downloaded. 
        Defaults to None.
    """
    # Check inputs
    if split == "val":
        split = "validation"
    
    if isinstance(path, str):
        path = pathlib.Path(path)

    ml_collection = load_secret(secret)
    if verbose:
        print("Downloading %s" % ml_collection["id"])

    # Download the dataset
    if split is None:
        _f = ("train", "test", "validation")
        [(path / x).mkdir(parents=True, exist_ok=True) for x in _f]
        for lf in _f:
            download_split(ml_collection, path, lf)
    elif split in ("validation", "train", "test"):
        (path / split).mkdir(parents=True, exist_ok=True)
        download_split(ml_collection, path, split)

    return path


def default_data_dir() -> pathlib.Path:
    """ Get the default data directory."""
    root = pathlib.Path.home() / ".mlstac/datasets/"
    root.mkdir(parents=True, exist_ok=True)

    return root


def load(
    secret: Union[str, pathlib.Path, dict],
    stream: bool = False,
    data_dir: Union[str, pathlib.Path] = None,
    split: Union[str, None] = "train",
    framework: str = "torch",
    device: str = "cpu",
    verbose: bool = True,
):

    if not stream:
        if data_dir is None:
            data_dir = default_data_dir()
        elif isinstance(data_dir, str):
            data_dir = pathlib.Path(data_dir)

        download(secret, data_dir, split, verbose)

        return LocalDataset(path=data_dir / split, framework=framework, device=device)

    else:
        ml_collection = load_secret(secret)

        # create a alway empty temporary file
        tempobj = NamedTemporaryFile(suffix=".safetensors")
        tempfile = pathlib.Path(tempobj.name)

        # obtain the url
        url = ml_collection["ml:%s" % split]["url"]
        if verbose:
            print("Connecting to %s" % ml_collection["id"])

        return StreamDataset(
            tempfile=tempfile, url=url, framework=framework, device=device
        )



from .datamodel import Catalog

# Create an ML-STAC Catalog
ml_catalog = Catalog(name="train", n_items=100)

# Create JSON Schema
# import json
# with open('mlstac-spec/catalog-spec/squema.json', 'w') as f:
#     json.dump(ml_catalog.model_json_schema(), f, indent=4)



from datetime import datetime

from ..catalog.dataclass import Catalog
from .dataclass import (Author, Authors, Collection, Extent, HyperLink,
                        License, Licenses, Reviewer, Reviewers, SpatialExtent,
                        TemporalExtent)

# Create an ML-STAC Catalog
ml_catalog = Collection(
    mlstac_version="0.1.0",
    name="cloudsen12",
    train=Catalog(name="train", n_items=100),
    val=Catalog(name="val", n_items=20),
    test=Catalog(name="test", n_items=20),
    authors=Authors(
        authors=[
            Author(
                name="Javier",
                orcid="https://orcid.org/0000-0001-0000-0000",
                organization="UPM",
            ),
            Author(
                name="Luis",
                orcid="https://orcid.org/0000-0002-0000-0000",
                organization="UPM",
            ),
            Author(
                name="Lola",
                orcid="https://orcid.org/0000-0003-0000-0000",
                organization="IPL",
            ),
        ],
        additional_comments=(
            "The authors agree to publish the dataset under "
            + "the ML-STAC specification."
        ),
        curated_by=HyperLink(link="https://github.com/csaybar"),
        reference=HyperLink(link="https://cloudsen12.github.io/"),
    ),
    licenses=Licenses(
        licenses=[
            License(
                name="CC-BY-4.0",
                link=HyperLink(link="https://creativecommons.org/licenses/by/4.0/"),
            )
        ],
        additional_comments=(
            "The authors agree to change the license from CC-BY-4.0 "
            + "to CC-BY-SA-4.0."
        ),
    ),
    split_strategy="spatial stratified random sampling",
    cv_task="multi-class semantic segmentation",
    sensor="sentinel-2 msi",
    bands=[
        "aerosol",
        "blue",
        "green",
        "red",
        "rededge1",
        "rededge2",
        "rededge3",
        "nir",
        "nir2",
        "watervapor",
        "cirrus",
        "swir1",
        "swir2",
    ],
    review=Reviewers(
        reviewers=[
            Reviewer(
                name="Javier",
                orcid="https://orcid.org/0000-0001-0000-0000",
                score=2,
                comments="The dataset is not good",
                issue_link=HyperLink(link="https://github.com/r-spatial/rgee/issues/1"),
            )
        ]
    ),
    extent=Extent(
        spatial=SpatialExtent(bbox=[-180, -90, 180, 90]),
        temporal=TemporalExtent(interval=[datetime(2019, 1, 1), datetime(2020, 1, 1)]),
    ),
    dtype_input="uint16",
    dtype_target="uint8",
    dtype_reference="uint8",
    size=100,
    keywords=["cloud-detection", "cloud-segmentation"],
    description=(
        "CloudSEN12 is a LARGE dataset (~1 TB) for cloud semantic "
        + "understanding that consists of 49,400 image patches (IP) "
        + "that  are evenly spread throughout all continents except "
        + "Antarctica. Each IP covers 5090 x 5090 meters and "
        + "contains data from Sentinel-2 levels 1C and 2A, hand-crafted "
        + "annotations of thick and thin clouds and cloud shadows, "
        + "Sentinel-1 Synthetic Aperture Radar (SAR), digital "
        + "elevation model, surface water occurrence, land cover "
        + "classes, and cloud mask results from six cutting-edge "
        + "cloud detection algorithms."
    ),
    extra={"github": "https://github.com/cloudsen12", "dummy": "dummy"},
    issue_link=HyperLink(link="https://github.com/r-spatial/rgee/issues/2"),
)

# Create JSON Schema
# import json
# with open('mlstac-spec/collection/squema.json', 'w') as f:
#     json.dump(ml_catalog.model_json_schema(), f, indent=4)



import io
import os
import re

from setuptools import find_packages, setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    name="mlstac",
    version="0.0.5",
    url="https://github.com/IPL-UV/ml-stac",
    license="GNU GPLv3",
    author="Cesar Aybar, Julio Contreras",
    author_email="cesar.aybar@uv.es, julio.contreras1@unmsm.edu.pe",
    description="A Common Language for EO Machine Learning Data",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",)),
    # remove all the example.py
    package_data={"mlstac": ["*example.py"]},
    install_requires=["pydantic>=2.4.2", "safetensors", "tqdm"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import gzip
import hashlib
import importlib
import pathlib
import re
from datetime import datetime
from tempfile import NamedTemporaryFile
from typing import Any, Optional

import pydantic

BASE_ALPH = tuple("123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")
BASE_DICT = dict((c, v) for v, c in enumerate(BASE_ALPH))
BASE_LEN = len(BASE_ALPH)


class SampleTensor(pydantic.BaseModel):
    # Fields -------------------------------------------------
    input: Any = None
    target: Any = None
    extra: Any = None
    tensor_framework: Optional[str] = None

    # Check Framework ----------------------------------------
    @classmethod
    def check_framework(cls, name: str) -> bool:
        try:
            importlib.import_module(name)
            return True
        except ModuleNotFoundError:
            return False

    @classmethod
    def check_is_a_tensor(cls, v: Any) -> bool:
        """ Check that the input is a valid tensor.

        Args:
            v (Any): The tensor to check. It can be a numpy, 
            torch, paddle, tensorflow or jax tensor.
        """

        np_ok = cls.check_framework("numpy")
        if np_ok:
            import numpy as np

            if type(v) == eval("np.ndarray"):
                cls._tensor_framework = "numpy"
                return True

        torch_ok = cls.check_framework("torch")
        if torch_ok:
            import torch

            if type(v) == eval("torch.Tensor"):
                cls._tensor_framework = "torch"
                return True

        jax_ok = cls.check_framework("jax")
        if jax_ok:
            import jax

            if type(v) == eval("jax.numpy.ndarray"):
                cls._tensor_framework = "jax"
                return True

        tensorflow_ok = cls.check_framework("tensorflow")
        if tensorflow_ok:
            import tensorflow

            if type(v) == eval("tensorflow.Tensor"):
                cls._tensor_framework = "tensorflow"
                return True

        paddle_ok = cls.check_framework("paddle")
        if paddle_ok:
            import paddle

            if type(v) == eval("paddle.Tensor"):
                cls._tensor_framework = "paddle"
                return True

        return False

    # Field Validators ---------------------------------------
    @pydantic.field_validator("input", "target", "extra")
    def check_target_and_extra(cls, v) -> None:
        if v is not None:
            if cls.check_is_a_tensor(v) is False:
                msg = "%s%s" % (
                    "The target or extra tensor must be None, ",
                    "or a numpy, torch, paddle or jax tensor.",
                )
                raise ValueError(msg)
        return v

    # Model Validators ---------------------------------------
    @pydantic.model_validator(mode="after")
    def check_model_none(self) -> "SampleTensor":
        if self.input is None:
            if self.target is None:
                if self.extra is None:
                    raise ValueError("The input, target and extra cannot all be None")
        return self

    @pydantic.model_validator(mode="after")
    def check_model_format(self) -> "SampleTensor":
        evaluation = "type(self.input)"
        if self.target is not None:
            evaluation = f"{evaluation} == type(self.target)"

        if self.extra is not None:
            evaluation = f"{evaluation} == type(self.extra)"

        if not eval(evaluation):
            raise ValueError("The input, target and extra must have the same format")
        self.tensor_framework = self._tensor_framework
        return self


class SampleMetadata(pydantic.BaseModel):

    # Fields -------------------------------------------------
    input: Optional[str] = None
    target: Optional[str] = None
    id: int
    geotransform: list
    crs: str
    start_datetime: datetime
    end_datetime: Optional[datetime] = None

    _safe_id: Optional[str] = pydantic.PrivateAttr()
    _safe_geotransform: Optional[str] = pydantic.PrivateAttr()
    _safe_start_datetime: Optional[str] = pydantic.PrivateAttr()
    _safe_end_datetime: Optional[str] = pydantic.PrivateAttr()

    # Field Validators ---------------------------------------
    @pydantic.field_validator("geotransform")
    def check_geotransform(cls, v: list) -> list:
        if len(v) != 6:
            raise ValueError("geotransform must contain 6 elements")

        if not all(isinstance(x, (int, float)) for x in v):
            raise ValueError("geotransform must contain only numbers")

        return v

    @pydantic.field_validator("crs")
    def check_crs(cls, v: str) -> str:
        regex_exp = re.compile(r"^(?:EPSG|ESRI|SR-ORG):[0-9]+$")

        if not regex_exp.match(v):
            raise ValueError("crs must be from the format: <authority>:<code>")

        return v

    # Model Validators ---------------------------------------
    @classmethod
    def base61_encode(cls, num: int) -> str:
        """ Encode a number in base61.

        Args:
            num (int): The number to encode.

        Returns:
            str: The encoded string.
        """

        if not num:
            return BASE_ALPH[0]

        encoding = ""
        while num:
            num, rem = divmod(num, BASE_LEN)
            encoding = BASE_ALPH[rem] + encoding
        return encoding

    @pydantic.model_validator(mode="after")
    def check_start_end_datetime(self) -> "SampleMetadata":
        if self.end_datetime is not None:
            if self.start_datetime > self.end_datetime:
                raise ValueError("end_datetime must be after start_datetime")

        # Set the safe metadata
        self._safe_id = self.base61_encode(self.id).zfill(7)
        self._safe_geotransform = " ,".join([str(x) for x in self.geotransform])
        self._safe_start_datetime = self.start_datetime.isoformat()
        if self.end_datetime is not None:
            self._safe_end_datetime = self.end_datetime.isoformat()
        else:
            self._safe_end_datetime = None

        return self


class Sample(pydantic.BaseModel):
    tensor: SampleTensor
    metadata: SampleMetadata
    _filename: Optional[str] = pydantic.PrivateAttr()

    def save(self, path: str, compress: bool = True) -> None:
        """Save an ML-STAC sample to disk.

        Args:
            path (str): The path to save the sample to.
            compress (bool, optional): Whether to compress the sample
                using gzip. Defaults to True.
        """

        path = pathlib.Path(path)

        if compress:
            tmp_file = NamedTemporaryFile()
            filename = pathlib.Path(tmp_file.name)
            gz_filename = path / f"{self.metadata._safe_id}.safetensors.gz"
            self._filename = gz_filename
        else:
            filename = path / f"{self.metadata._safe_id}.safetensors"
            self._filename = filename

        metadata = {
            "input": self.metadata.input,
            "target": self.metadata.target,
            "id": self.metadata._safe_id,
            "geotransform": self.metadata._safe_geotransform,
            "crs": self.metadata.crs,
            "start_datetime": self.metadata._safe_start_datetime,
            "end_datetime": self.metadata._safe_end_datetime,
        }
        tensors = self.tensor.model_dump()

        metadata = {k: v for k, v in metadata.items() if v is not None}
        tensors = {k: v for k, v in tensors.items() if v is not None}
        tensor_framework = tensors.pop("tensor_framework")

        if tensor_framework == "numpy":
            import safetensors.numpy

            safetensors.numpy.save_file(tensors, metadata=metadata, filename=filename)
        elif tensor_framework == "torch":
            import safetensors.torch

            safetensors.torch.save_file(tensors, metadata=metadata, filename=filename)
        elif tensor_framework == "paddle":
            import safetensors.paddle

            safetensors.paddle.save_file(tensors, metadata=metadata, filename=filename)
        elif tensor_framework == "tensorflow":
            import safetensors.tensorflow

            safetensors.tensorflow.save_file(
                tensors, metadata=metadata, filename=filename
            )
        elif tensor_framework == "jax":
            import safetensors.jax

            safetensors.jax.save_file(tensors, metadata=metadata, filename=filename)

        if compress:
            with open(filename, "rb") as f_in:
                with gzip.open(gz_filename, "wb") as f_out:
                    f_out.writelines(f_in)
            tmp_file.close()

    def checksum(self) -> None:
        filename = self._filename

        if not filename.exists():
            raise FileNotFoundError(f"{filename} does not found")

        # Checksum
        with open(filename, "rb") as f:
            checksum = hashlib.md5(f.read()).hexdigest()

        return checksum



import pathlib
import mlstac
import numpy as np

secret = "https://huggingface.co/datasets/jfloresf/burn-cems-1/resolve/main/main.json"
train_db = mlstac.load(
    secret=secret,
    framework="numpy",
    stream=False,
    device="cpu",
    data_dir=pathlib.Path("/home/gonzalo/deno")
)


batch = train_db[0]
input = batch["input"]
target = batch["target"]
extra = batch["extra"]

import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 3)
ax[0].imshow(np.moveaxis(input[[3, 2, 1], :, :], 0, -1)*3)
ax[1].imshow(target[0, :, :])
ax[2].imshow(np.moveaxis(input[[3, 2, 1], :, :], 0, -1)*3)
plt.show()

import time
from tqdm import tqdm
with tqdm(total=100) as pbar:
    for i in range(10):
        time.sleep(0.1)
        pbar.update(10)
pbar.close()


# to pypi
!python setup.py sdist bdist_wheel
!twine upload dist/*


from datetime import datetime

import numpy as np

from .datamodel import Sample, SampleMetadata, SampleTensor

# Set the ML-STAC sample metadata
sample_metadata = SampleMetadata(
    id=40000,
    geotransform=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    crs="SR-ORG:4326",
    start_datetime=datetime(2021, 1, 1, 0, 0, 0),
    end_datetime=None,
)


# Set the ML-STAC sample tensors
sample_tensors = SampleTensor(
    input=np.array([1, 2, 3]), target=np.array([1, 2, 3]), extra=None
)

# Set the ML-STAC item
sample = Sample(metadata=sample_metadata, tensor=sample_tensors)
sample.save(path="/home/gonzalo/")
sample.checksum()

# Write the squema
with open("mlstac/sample/squema.json", "w") as f:
    f.write(sample.model_schema_json(indent=2))
